"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FfmpegProcess = void 0;
const execa_1 = __importDefault(require("execa"));
const dgram_1 = require("dgram");
class FfmpegProcess {
    constructor(delegate, sessionId, command, returnPort, callback) {
        this.command = command.join(" ");
        this.debug = delegate.platform.debug.bind(delegate.platform);
        this.delegate = delegate;
        this.log = delegate.platform.log;
        this.name = delegate.name.bind(delegate.protectCamera);
        this.sessionId = sessionId;
        // Toggle FFmpeg logging, if configured.
        this.isVerbose = this.delegate.platform.verboseFfmpeg;
        // Create the return port for FFmpeg, if requested to do so. The only time we don't do this is when we're standing up
        // a two-way audio stream - in that case, the audio work is done through RtpSplitter and not here.
        if (returnPort) {
            this.createSocket(returnPort);
        }
        void this.startFfmpeg(command, callback);
    }
    // Create the port for FFmpeg to send data through.
    createSocket(portInfo) {
        const socket = dgram_1.createSocket(portInfo.addressVersion === "ipv6" ? "udp6" : "udp4");
        // Cleanup after ourselves when the socket closes.
        socket.on("close", () => {
            if (this.streamTimeout) {
                clearTimeout(this.streamTimeout);
            }
        });
        // Handle potential network errors.
        socket.on("error", (error) => {
            this.log.error("%s: Socket error: %s.", this.name(), error.name);
            this.delegate.stopStream(this.sessionId);
        });
        // Manage our video streams in case we haven't received a stop request, but we're in fact dead zombies.
        socket.on("message", () => {
            // Clear our last canary.
            if (this.streamTimeout) {
                clearTimeout(this.streamTimeout);
            }
            // Set our new canary.
            this.streamTimeout = setTimeout(() => {
                this.debug("%s: video stream appears to be inactive for 5 seconds. Stopping stream.", this.name());
                this.delegate.controller.forceStopStreamingSession(this.sessionId);
                this.delegate.stopStream(this.sessionId);
            }, 5000);
        });
        socket.bind(portInfo.port);
    }
    // Start our FFmpeg process.
    async startFfmpeg(ffmpegCommandLine, callback) {
        var _a, _b, _c, _d;
        // See if we should display ffmpeg command output.
        let hasLogging = false;
        // Track if we've started receiving data.
        let started = false;
        // If we've got a loglevel specified, ensure we display it.
        if (ffmpegCommandLine.indexOf("-loglevel") !== -1) {
            hasLogging = true;
        }
        if (hasLogging || this.isVerbose || this.delegate.platform.config.debugAll) {
            this.log.info("%s: ffmpeg command: %s %s", this.name(), this.delegate.videoProcessor, ffmpegCommandLine.join(" "));
        }
        else {
            this.debug("%s: ffmpeg command: %s %s", this.name(), this.delegate.videoProcessor, ffmpegCommandLine.join(" "));
        }
        // Prepare the command line we want to execute.
        this.process = execa_1.default(this.delegate.videoProcessor, ffmpegCommandLine);
        // Handle errors on stdin.
        (_a = this.process.stdin) === null || _a === void 0 ? void 0 : _a.on("error", (error) => {
            if (!error.message.includes("EPIPE")) {
                this.log.error("%s: FFmpeg error: %s.", this.name(), error.message);
            }
        });
        // Handle logging output that gets sent to stderr.
        (_b = this.process.stderr) === null || _b === void 0 ? void 0 : _b.on("data", (data) => {
            if (!started) {
                started = true;
                this.debug("%s: Received the first frame.", this.name());
                // Always remember to execute the callback once we're setup to let homebridge know we're streaming.
                if (callback) {
                    callback();
                }
            }
            // Debugging and additional logging, if requested.
            if (hasLogging || this.isVerbose || this.delegate.platform.config.debugAll) {
                data.toString().split(/\n/).forEach((line) => {
                    this.log.info(line);
                });
            }
        });
        try {
            // Execute the command line.
            await this.process;
        }
        catch (error) {
            // You might think this should be ExecaError, but ExecaError is a type, not a class, and instanceof
            // only operates on classes.
            if (!(error instanceof Error)) {
                this.log.error("Unknown error received while attempting to start FFmpeg: %s.", error);
                return;
            }
            // Recast our error object as an ExecaError.
            const execError = error;
            // Some utilities to streamline things.
            const logPrefix = this.name() + ": FFmpeg process ended ";
            const code = (_c = execError.exitCode) !== null && _c !== void 0 ? _c : null;
            const signal = (_d = execError.signal) !== null && _d !== void 0 ? _d : null;
            // We asked FFmpeg to stop.
            if (execError.isCanceled) {
                this.debug(logPrefix + "(Expected).");
                return;
            }
            // FFmpeg ended for another reason.
            const errorMessage = logPrefix + "(Error)." + (code === null ? "" : " Exit code: " + code.toString() + ".") + (signal === null ? "" : " Signal: " + signal + ".");
            this.log.error("%s: %s", this.name(), execError.message);
            // Stop the stream.
            this.delegate.stopStream(this.sessionId);
            // Temporarily increase logging verbosity.
            this.delegate.setVerboseFfmpeg();
            // Let homebridge know what happened and stop the stream if we've already started.
            if (!started && callback) {
                callback(new Error(errorMessage));
                return;
            }
            this.delegate.controller.forceStopStreamingSession(this.sessionId);
        }
    }
    // Cleanup after we're done.
    stop() {
        // Cancel our process.
        this.process.cancel();
    }
    // Grab the standard input.
    getStdin() {
        return this.process.stdin;
    }
    // Grab the standard output.
    getStdout() {
        return this.process.stdout;
    }
    // Validate whether or not we have a specific codec available to us in FFmpeg.
    static async codecEnabled(videoProcessor, codec) {
        const output = await execa_1.default(videoProcessor, ["-codecs"]);
        return output.stdout.includes(codec);
    }
}
exports.FfmpegProcess = FfmpegProcess;
//# sourceMappingURL=protect-ffmpeg.js.map