"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RtpUtils = exports.RtpDemuxer = void 0;
const settings_1 = require("./settings");
const dgram_1 = require("dgram");
const get_port_1 = __importDefault(require("get-port"));
/*
 * Here's the problem this class solves: FFmpeg doesn't support multiplexing RTP and RTCP data on a single UDP port (RFC 5761).
 * If it did, we wouldn't need this workaround for HomeKit compatibility, which does multiplex RTP and RTCP over a single UDP port.
 *
 * This class inspects all packets coming in from inputPort and demultiplexes RTP and RTCP traffic to rtpPort and rtcpPort, respectively.
 *
 * Credit to @dgreif and @brandawg93 who graciously shared their code as a starting point, and their collaboration
 * in answering the questions needed to bring all this together. A special thank you to @Sunoo for the many hours of
 * discussion and brainstorming on this and other topics.
 */
class RtpDemuxer {
    // Create an instance of RtpDemuxer.
    constructor(streamingDelegate, ipFamily, inputPort, rtcpPort, rtpPort) {
        this.debug = streamingDelegate.platform.debug.bind(streamingDelegate.platform);
        this.delegate = streamingDelegate;
        this.log = streamingDelegate.log;
        this.inputPort = inputPort;
        this.socket = dgram_1.createSocket(ipFamily === "ipv6" ? "udp6" : "udp4");
        // Catch errors when they happen on our demuxer.
        this.socket.on("error", (error) => {
            this.log.error("%s: RtpDemuxer Error: %s", this.delegate.protectCamera.name(), error);
            this.socket.close();
        });
        // Split the message into RTP and RTCP packets.
        this.socket.on("message", (msg) => {
            // Send RTP packets to the RTP port.
            if (this.isRtpMessage(msg)) {
                this.socket.send(msg, rtpPort);
            }
            else {
                // Save this RTCP message for heartbeat purposes for the RTP port. This works because RTCP packets will be ignored
                // by ffmpeg on the RTP port, effectively providing a heartbeat to ensure FFmpeg doesn't timeout if there's an
                // extended delay between data transmission.
                this.heartbeatMsg = Buffer.from(msg);
                // Clear the old heartbeat timer.
                clearTimeout(this.heartbeatTimer);
                this.heartbeat(rtpPort);
                // RTCP control packets should go to the RTCP port.
                this.socket.send(msg, rtcpPort);
            }
        });
        this.debug("%s: Creating an RtpDemuxer instance - inbound port: %s, RTCP port: %s, RTP port: %s.", this.delegate.protectCamera.name(), this.inputPort, rtcpPort, rtpPort);
        // Take the socket live.
        this.socket.bind(this.inputPort);
    }
    // Send a regular heartbeat to FFmpeg to ensure the pipe remains open and the process alive.
    heartbeat(port) {
        // Clear the old heartbeat timer.
        clearTimeout(this.heartbeatTimer);
        // Send a heartbeat to FFmpeg every few seconds to keep things open. FFmpeg has a five-second timeout
        // in reading input, and we want to be comfortably within the margin for error to ensure the process
        // continues to run.
        this.heartbeatTimer = setTimeout(() => {
            this.debug("Sending ffmpeg a heartbeat.");
            this.socket.send(this.heartbeatMsg, port);
            this.heartbeat(port);
        }, settings_1.PROTECT_TWOWAY_HEARTBEAT_INTERVAL * 1000);
    }
    // Close the socket and cleanup.
    close() {
        this.debug("%s: Closing the RtpDemuxer instance on port %s.", this.delegate.protectCamera.name(), this.inputPort);
        clearTimeout(this.heartbeatTimer);
        this.socket.close();
    }
    // Retrieve the payload information from a packet to discern what the packet payload is.
    getPayloadType(message) {
        return message.readUInt8(1) & 0x7f;
    }
    // Return whether or not a packet is RTP (or not).
    isRtpMessage(message) {
        const payloadType = this.getPayloadType(message);
        return (payloadType > 90) || (payloadType === 0);
    }
}
exports.RtpDemuxer = RtpDemuxer;
// RTP-related utilities.
class RtpUtils {
    // Reserve consecutive ports for use with FFmpeg. FFmpeg currently lacks the ability to specify both the RTP
    // and RTCP ports. It always assumes, by convention, that when you specify an RTP port, the RTCP port is the
    // RTP port + 1. In order to work around that challenge, we need to always ensure that when we reserve multiple
    // ports for RTP (primarily for two-way audio) that we we are reserving consecutive ports only.
    static async reservePorts(count = 1) {
        // Get the first port.
        const port = await get_port_1.default();
        const ports = [port];
        // If we're requesting additional consecutive ports, keep searching until they're found.
        for (let i = 1; i < count; i++) {
            const targetConsecutivePort = port + i;
            // We need to await here in order to determine whether we need to keep going.
            // eslint-disable-next-line no-await-in-loop
            const openPort = await get_port_1.default({ port: targetConsecutivePort });
            // Unable to reserve the next consecutive port. Roll the dice again and hope for the best.
            if (openPort !== targetConsecutivePort) {
                return this.reservePorts(count);
            }
            ports.push(openPort);
        }
        return ports;
    }
}
exports.RtpUtils = RtpUtils;
//# sourceMappingURL=protect-rtp.js.map